// Lightray3D (C) 2002-2006 Tzivenis Sakis

#ifndef PRIMITIVES_H
#define PRIMITIVES_H

#include "cmesh.h"



//---------------------------------------------------------------------------------------------
// Base class
//---------------------------------------------------------------------------------------------
class COREAPI cprimitive
{
public:
	cprimitive();
	virtual ~cprimitive(){}

	virtual void Clear(){}//=0;
	virtual BOOL BuildMesh(cmesh&){return FALSE;}//=0;
};


//---------------------------------------------------------------------------------------------
// Plane
//---------------------------------------------------------------------------------------------
class COREAPI cplane : public cprimitive
{
public:
	cplane();
	~cplane();

	BOOL Create(cmesh&,
				float size,float scalew,float scaleh,
				int nhsections=0,int nvsections=0);
	BOOL Create(cmesh&,
				float width,float height,
				int nhsections=0,int nvsections=0);


	void SetWidth(float);
	void SetHeight(float);
	void SetHSections(int);
	void SetVSections(int);

	float GetWidth()const {return width;}
	float GetHeight()const {return height;}
	int GetHSections()const {return nhsections;}
	int GetVSections()const {return nvsections;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float width,height;
	int nhsections,nvsections;
};



//---------------------------------------------------------------------------------------------
// Disc
//---------------------------------------------------------------------------------------------
class COREAPI cdisc : public cprimitive
{
public:
	cdisc();
	~cdisc();

	BOOL Create(cmesh&,float radius,int nsides,int nrings=0);


	void SetRadius(float);
	void SetSides(int);
	void SetRings(int);

	float GetRadius()const {return radius;}
	int GetSides()const {return sides;}
	int GetRings()const {return rings;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float radius;
	int sides,rings;
};





//---------------------------------------------------------------------------------------------
// Cube
//---------------------------------------------------------------------------------------------
class COREAPI ccube : public cprimitive
{
public:
	ccube();
	~ccube();

	BOOL Create(cmesh&,
				float width,float height,float length,
				int wsections,int hsections,int lsections,
				BOOL bBottomPivot=FALSE);//(ver.>=1.3.6 update1)


	void SetWidth(float);
	void SetHeight(float);
	void SetLength(float);
	void SetWidthSections(int);
	void SetHeightSections(int);
	void SetLengthSections(int);

	float GetWidth()const {return width;}
	float GetHeight()const {return height;}
	float GetLength()const {return length;}
	int GetWidthSections()const {return widthsections;}
	int GetHeightSections()const {return heightsections;}
	int GetLengthSections()const {return lengthsections;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float width,height,length;
	int widthsections,heightsections,lengthsections;
	BOOL bBottomPiv;
};

//---------------------------------------------------------------------------------------------
// Sphere
//---------------------------------------------------------------------------------------------
class COREAPI csphere : public cprimitive
{
public:
	csphere();
	~csphere();

	BOOL Create(cmesh&,
				float radius, int num_rings, int num_sections,
				float scalex, float scaley, float scalez);
	BOOL Create(cmesh&,
				float radiusx, float radiusy, float radiusz,
				int num_rings, int num_sections);


	void SetRadius(float);
	void SetRadiusX(float);
	void SetRadiusY(float);
	void SetRadiusZ(float);
	void SetRings(int);
	void SetSections(int);

	float GetRadiusX()const {return rX;}
	float GetRadiusY()const {return rY;}
	float GetRadiusZ()const {return rZ;}
	int GetRings()const {return nrings;}
	int GetSections()const {return nsections;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float rX,rY,rZ;
	int nrings,nsections;
};


//---------------------------------------------------------------------------------------------
// Cylinder
//---------------------------------------------------------------------------------------------
class COREAPI ccylinder : public cprimitive
{
public:
	ccylinder();
	~ccylinder();

	BOOL Create(cmesh&,
				float bottomradius,float topradius,
				float height,int nsides,int nsections,int ncapsections,
				BOOL bBottomPivot=FALSE);//(ver.>=1.3.6 update1)


	void SetBottomRadius(float);
	void SetTopRadius(float);
	void SetHeight(float);
	void SetSides(int);
	void SetSections(int);
	void SetCapSections(int);

	float GetBottomRadius()const {return bradius;}
	float GetTopRadius()const {return tradius;}
	float GetHeight()const {return height;}
	int GetSides()const {return sides;}
	int GetSections()const {return sections;}
	int GetCapSections()const {return capsections;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float bradius,tradius,height;
	int sides,sections,capsections;
	BOOL bBottomPiv;
};

//---------------------------------------------------------------------------------------------
// Cone
//---------------------------------------------------------------------------------------------
class COREAPI ccone : public cprimitive
{
public:
	ccone();
	~ccone();

	BOOL Create(cmesh&,
				float radius,float height,int nsides,int nsections,int ncapsections,
				BOOL bBottomPivot=FALSE);//(ver.>=1.3.6 update1)

	void SetRadius(float);
	void SetHeight(float);
	void SetSides(int);
	void SetSections(int);
	void SetCapSections(int);

	float GetRadius()const {return radius;}
	float GetHeight()const {return height;}
	int GetSides()const {return sides;}
	int GetSections()const {return sections;}
	int GetCapSections()const {return capsections;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float radius,height;
	int sides,sections,capsections;
	BOOL bBottomPiv;
};

//---------------------------------------------------------------------------------------------
// Tube
//---------------------------------------------------------------------------------------------
class COREAPI ctube : public cprimitive
{
public:
	ctube();
	~ctube();

	BOOL Create(cmesh&,
				float innerradius,float outerradius,
				float height,int nsides,int nsections,
				BOOL bBottomPivot=FALSE);//(ver.>=1.3.6 update1)


	void SetInnerRadius(float);
	void SetOuterRadius(float);
	void SetHeight(float);
	void SetSides(int);
	void SetSections(int);

	float GetInnerRadius()const {return inradius;}
	float GetOuterRadius()const {return outradius;}
	float GetHeight()const {return height;}
	int GetSides()const {return sides;}
	int GetSections()const {return sections;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float inradius,outradius,height;
	int sides,sections;
	BOOL bBottomPiv;
};


//---------------------------------------------------------------------------------------------
// Donut
//---------------------------------------------------------------------------------------------
class COREAPI cdonut : public cprimitive
{
public:
	cdonut();
	~cdonut();

	BOOL Create(cmesh&,
				float innerradius,float outerradius,
				int nsides,int nsections);


	void SetInnerRadius(float);
	void SetOuterRadius(float);
	void SetSides(int);
	void SetSections(int);

	float GetInnerRadius()const {return inradius;}
	float GetOuterRadius()const {return outradius;}
	int GetSides()const {return sides;}
	int GetSections()const {return sections;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float inradius,outradius;
	int sides,sections;
};


//---------------------------------------------------------------------------------------------
// Geosphere
// (ver.>=1.3.1):Added in core.dll
//---------------------------------------------------------------------------------------------
#define GEO_TYPE_TETRA		0
#define GEO_TYPE_OCTA		1
#define GEO_TYPE_ICOSA		2

class COREAPI cgeosphere : public cprimitive
{
public:
	cgeosphere();
	~cgeosphere();

	BOOL Create(cmesh&, int type, float radius, int num_segs);


	void SetRadius(float);
	void SetSegments(int);
	void SetType(int);

	float GetRadius()const {return radius;}
	int GetSegments()const {return nsegments;}
	int GetType()const {return basetype;}

	//From cprimitive
	void Clear();
	BOOL BuildMesh(cmesh&);

protected:
	float radius;
	int nsegments;
	int basetype;
};

#endif //PRIMITIVES_H