// Lightray3D (C) 2002-2006 Tzivenis Sakis

#ifndef CMATHLIB_H
#define CMATHLIB_H

#include <math.h>

//NOTE: All angle parameters/return values must be/are in degrees

#pragma warning(disable : 4244)		// conversion from int to float
#pragma warning(disable : 4305)     // truncate from double to float

//#define FLOAT_SMALL		(float)1.0e-6// min positive value
//#define FLOAT_BIG			(float)1.0e+17// max value
#define FLOAT_SMALL			(float)1.0e-10// min positive value
#define FLOAT_BIG			(float)1.0e+20// max value


// useful definitions
#undef  PI
#define PI				3.14159265358979323846	// Pi
#undef  PI2
#define PI2				6.28318530717958623200	// Pi*2
#define PIDIV2			1.57079632679489655800	// Pi/2
#define PIDIV4			0.78539816339744827900	// Pi/4
#define PIINV			0.78539816339744827900	// 1/Pi
#define DEGTORAD		0.01745329251994329547  // Degrees to Radians
#define RADTODEG		57.29577951308232286465 // Radians to Degrees

// macros to convert from degrees to radians and radians to degrees
#define DEG2RAD(val) (val*DEGTORAD)
#define RAD2DEG(val) (val*RADTODEG)

// axes
#undef AXIS_X
#define AXIS_X		(1<<0)
#undef AXIS_Y
#define AXIS_Y		(1<<1)
#undef AXIS_Z
#define AXIS_Z		(1<<2)


#define INTERPOLATE(l,h,t) ((l) + ((h) - (l)) * (t))
#define RINT(in)	floor(in + 0.5)
#define SNAPFLOAT(v) (float)floor(v + 0.5)
#define SNAPTOGRID(v,g) (float)(floor(v / g + 0.5) * g)


//forward declarations
class cmatrix4;

//------------------------------------------------------------------------------------------
// Point2
//------------------------------------------------------------------------------------------
class COREAPI cpoint2
{
	friend COREAPI cpoint2 MakePt2(float X, float Y);

public:
	float x,y;

	cpoint2(){ x = 0.0f; y = 0.0f; }
	cpoint2(const cpoint2& pt) { x = pt.x; y = pt.y; } 
	cpoint2(float X, float Y)  { x = X; y = Y; }
	cpoint2(int X, int Y) { x = (float)X; y = (float)Y; }
	cpoint2(float pt[2]) { x = pt[0]; y = pt[1]; }

	~cpoint2(){}//(ver.>=1.3.4):needed for Rayscript

	inline float Length() const;
	void Normalize();//in place

	//(ver.>=1.3.6):Bounding Box
	inline void AddMins(const cpoint2 &p);
	inline void AddMaxs(const cpoint2 &p);

	// Access operators
	float&  operator[](int i) { return (&x)[i]; }
	const float& operator[](int i) const { return (&x)[i]; }

	// Conversion function
	operator float*() { return (float*)(&x); }

	// Unary operators
	cpoint2 operator-() const { return(MakePt2(-x,-y)); } 
	cpoint2 operator+() const { return *this; }

	// Assignment operators
	inline cpoint2& operator-=(const cpoint2&);
	inline cpoint2& operator+=(const cpoint2&);
	inline cpoint2& operator/=(const cpoint2&);
	inline cpoint2& operator*=(const cpoint2&);
	inline cpoint2& operator-=(float);
	inline cpoint2& operator+=(float);
	inline cpoint2& operator/=(float);
	inline cpoint2& operator*=(float); 

	// Binary operators
	inline  cpoint2 operator-(const cpoint2&) const;
	inline  cpoint2 operator+(const cpoint2&) const;
	inline  cpoint2 operator/(const cpoint2&) const;
	inline  cpoint2 operator*(const cpoint2&) const;
	inline cpoint2 operator-(float) const;
	inline cpoint2 operator+(float) const;
	inline cpoint2 operator/(float) const;
	inline cpoint2 operator*(float) const;

	// Comparison operators
	inline int operator==(const cpoint2& p) const { return ((p.x==x)&&(p.y==y)); }
	inline int operator!=(const cpoint2& p) const { return ((p.x!=x)||(p.y!=y)); }

	//Max-Min component
	inline float Max() const
	{
		if(fabs(x)>=fabs(y)) return x;
		return y;
	}
	inline float Min() const
	{
		if(fabs(x)<=fabs(y)) return x;
		return y;
	}

	//Max-Min axis
	inline int MaxAxis() const
	{
		if(fabs(x)>=fabs(y)) return AXIS_X;
		return AXIS_Y;
	}
	inline int MinAxis() const
	{
		if(fabs(x)<=fabs(y)) return AXIS_X;
		return AXIS_Y;
	}

	//Interpolation
	inline cpoint2& Inter(const cpoint2& b,float t);
};

COREAPI cpoint2 Normalize(const cpoint2&);
COREAPI BOOL Compare(const cpoint2& a, const cpoint2& b, float tolerance);//tolerance >=0
COREAPI cpoint2 Inter(const cpoint2& a, const cpoint2& b, float t);


inline float cpoint2::Length() const
{
	return (float)sqrt(x*x+y*y);
}

//(ver.>=1.3.6):Bounding Box
inline void cpoint2::AddMins(const cpoint2 &p)
{
	if(p.x < x) x = p.x;
	if(p.y < y) y = p.y;
}

//(ver.>=1.3.6):Bounding Box
inline void cpoint2::AddMaxs(const cpoint2 &p)
{
	if(p.x > x) x = p.x;
	if(p.y > y) y = p.y;
}

inline cpoint2& cpoint2::operator-=(const cpoint2& a)
{	
	x -= a.x;	y -= a.y;
	return *this;
}

inline cpoint2& cpoint2::operator+=(const cpoint2& a)
{
	x += a.x;	y += a.y;
	return *this;
}

inline cpoint2& cpoint2::operator/=(const cpoint2& a)
{ 
	x /= a.x;	y /= a.y;
	return *this; 
}

inline cpoint2& cpoint2::operator*=(const cpoint2& a)
{ 
	x *= a.x;	y *= a.y;
	return *this; 
}

inline cpoint2& cpoint2::operator-=(float f)
{
	x -= f;   y -= f;
	return *this;
}

inline cpoint2& cpoint2::operator+=(float f)
{
	x += f;   y += f;
	return *this;
}

inline cpoint2& cpoint2::operator/=(float f)
{
	x /= f;   y /= f;
	return *this;
}

inline cpoint2& cpoint2::operator*=(float f)
{
	x *= f;   y *= f;
	return *this;
}

inline cpoint2 cpoint2::operator-(const cpoint2& b) const
{
	return(MakePt2(x-b.x,y-b.y));
}

inline cpoint2 cpoint2::operator+(const cpoint2& b) const
{
	return(MakePt2(x+b.x,y+b.y));
}

inline cpoint2 cpoint2::operator/(const cpoint2& b) const
{
	return MakePt2(x/b.x,y/b.y);
}

inline cpoint2 cpoint2::operator*(const cpoint2& b) const
{
	return MakePt2(x*b.x, y*b.y);	
}

inline cpoint2 cpoint2::operator-(float f) const
{
	return(MakePt2(x-f, y-f));
}

inline cpoint2 cpoint2::operator+(float f) const
{
	return(MakePt2(x+f, y+f));
}

inline cpoint2 cpoint2::operator/(float f) const
{
	return(MakePt2(x/f, y/f));
}

inline cpoint2 cpoint2::operator*(float f) const
{
	return(MakePt2(x*f, y*f));
}

inline float CrossProd(const cpoint2& a, const cpoint2& b)
{
	return (a.x*b.y - a.y*b.x);
}

inline float DotProd(const cpoint2& a, const cpoint2& b)
{
	return(a.x*b.x + a.y*b.y);
}

inline cpoint2& cpoint2::Inter(const cpoint2& b, float t)
{	
	x = INTERPOLATE(x, b.x, t);
	y = INTERPOLATE(y, b.y, t);
	return *this;
}



//------------------------------------------------------------------------------------------
// Point3
//------------------------------------------------------------------------------------------
class COREAPI cpoint3
{
	friend COREAPI cpoint3 MakePt3(float X, float Y, float Z);

public:
	float x,y,z;

	cpoint3(){ x = 0.0f; y = 0.0f; z = 0.0f; }
	cpoint3(const cpoint3& pt) { x = pt.x; y = pt.y; z = pt.z; } 
	cpoint3(float X, float Y, float Z)  { x = X; y = Y; z = Z; }
	cpoint3(int X, int Y, int Z) { x = (float)X; y = (float)Y; z = (float)Z; }
	cpoint3(float pt[3]) { x = pt[0]; y = pt[1]; z = pt[2]; }

	~cpoint3(){}//(ver.>=1.3.4):needed for Rayscript

	inline float Length() const;
	void Normalize();//in place

	//(ver.>=1.3.6):Bounding Box
	inline void AddMins(const cpoint3 &p);
	inline void AddMaxs(const cpoint3 &p);

	// Access operators
	float&  operator[](int i) { return (&x)[i]; }
	const float& operator[](int i) const { return (&x)[i]; }

	// Conversion function
	operator float*() { return (float*)(&x); }

	// Unary operators
	cpoint3 operator-() const { return(MakePt3(-x,-y,-z)); } 
	cpoint3 operator+() const { return *this; }

	// Assignment operators
	inline cpoint3& operator-=(const cpoint3&);
	inline cpoint3& operator+=(const cpoint3&);
	inline cpoint3& operator/=(const cpoint3&);
	inline cpoint3& operator*=(const cpoint3&);
	inline cpoint3& operator-=(float);
	inline cpoint3& operator+=(float);
	inline cpoint3& operator/=(float);
	inline cpoint3& operator*=(float); 

	// Binary operators
	inline  cpoint3 operator-(const cpoint3&) const;
	inline  cpoint3 operator+(const cpoint3&) const;
	inline  cpoint3 operator/(const cpoint3&) const;
	inline  cpoint3 operator*(const cpoint3&) const;
	inline cpoint3 operator-(float) const;
	inline cpoint3 operator+(float) const;
	inline cpoint3 operator/(float) const;
	inline cpoint3 operator*(float) const;
	
	cpoint3 operator*(const cmatrix4& M) const;//(ver.>=1.3.4):Moved here.

	// Comparison operators
	inline int operator==(const cpoint3& p) const { return ((p.x==x)&&(p.y==y)&&(p.z==z)); }
	inline int operator!=(const cpoint3& p) const { return ((p.x!=x)||(p.y!=y)||(p.z!=z)); }

	//Max-Min component
	inline float Max() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z);
		if(fX>=fY)
		{
			if(fX>=fZ) return x;
			return z;
		}
		if(fY>=fZ) return y;
		return z;
	}
	inline float Min() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z);
		if(fX<=fY)
		{
			if(fX<=fZ) return x;
			return z;
		}
		if(fY<=fZ) return y;
		return z;
	}

	//Max-Min axis
	inline int MaxAxis() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z);
		if(fX>=fY)
		{
			if(fX>=fZ) return AXIS_X;
			return AXIS_Z;
		}
		if(fY>=fZ) return AXIS_Y;
		return AXIS_Z;
	}
	inline int MinAxis() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z);
		if(fX<=fY)
		{
			if(fX<=fZ) return AXIS_X;
			return AXIS_Z;
		}
		if(fY<=fZ) return AXIS_Y;
		return AXIS_Z;
	}

	//Interpolation
	inline cpoint3& Inter(const cpoint3& b, float t);
};

COREAPI cpoint3 Normalize(const cpoint3&);
COREAPI cpoint3 CrossProd(const cpoint3& a, const cpoint3& b);
COREAPI BOOL Compare(const cpoint3& a, const cpoint3& b,float tolerance);//tolerance >=0
COREAPI cpoint3 Inter(const cpoint3& a, const cpoint3& b, float t);



inline float cpoint3::Length() const
{
	return (float)sqrt(x*x+y*y+z*z);
}

//(ver.>=1.3.6):Bounding Box
inline void cpoint3::AddMins(const cpoint3 &p)
{
	if(p.x < x) x = p.x;
	if(p.y < y) y = p.y;
	if(p.z < z) z = p.z;
}

//(ver.>=1.3.6):Bounding Box
inline void cpoint3::AddMaxs(const cpoint3 &p)
{
	if(p.x > x) x = p.x;
	if(p.y > y) y = p.y;
	if(p.z > z) z = p.z;
}

inline cpoint3& cpoint3::operator-=(const cpoint3& a)
{	
	x -= a.x;	y -= a.y;	z -= a.z;
	return *this;
}

inline cpoint3& cpoint3::operator+=(const cpoint3& a)
{
	x += a.x;	y += a.y;	z += a.z;
	return *this;
}

inline cpoint3& cpoint3::operator/=(const cpoint3& a)
{ 
	x /= a.x;	y /= a.y;	z /= a.z;	
	return *this; 
}

inline cpoint3& cpoint3::operator*=(const cpoint3& a)
{ 
	x *= a.x;	y *= a.y;	z *= a.z;	
	return *this; 
}

inline cpoint3& cpoint3::operator-=(float f)
{
	x -= f;   y -= f;	z -= f;
	return *this;
}

inline cpoint3& cpoint3::operator+=(float f)
{
	x += f;   y += f;	z += f;
	return *this;
}

inline cpoint3& cpoint3::operator/=(float f)
{
	x /= f;   y /= f;	z /= f;
	return *this;
}

inline cpoint3& cpoint3::operator*=(float f)
{
	x *= f;   y *= f;	z *= f;
	return *this;
}

inline cpoint3 cpoint3::operator-(const cpoint3& b) const
{
	return(MakePt3(x-b.x,y-b.y,z-b.z));
}

inline cpoint3 cpoint3::operator+(const cpoint3& b) const
{
	return(MakePt3(x+b.x,y+b.y,z+b.z));
}

inline cpoint3 cpoint3::operator/(const cpoint3& b) const
{
	return MakePt3(x/b.x,y/b.y,z/b.z);
}

inline cpoint3 cpoint3::operator*(const cpoint3& b) const
{
	return MakePt3(x*b.x, y*b.y, z*b.z);	
}

inline cpoint3 cpoint3::operator-(float f) const
{
	return(MakePt3(x-f, y-f, z-f));
}

inline cpoint3 cpoint3::operator+(float f) const
{
	return(MakePt3(x+f, y+f, z+f));
}

inline cpoint3 cpoint3::operator/(float f) const
{
	return(MakePt3(x/f, y/f, z/f));
}

inline cpoint3 cpoint3::operator*(float f) const
{
	return(MakePt3(x*f, y*f, z*f));
}

inline cpoint3& cpoint3::Inter(const cpoint3& b,float t)
{	
	x = INTERPOLATE(x, b.x, t);
	y = INTERPOLATE(y, b.y, t);
	z = INTERPOLATE(z, b.z, t);
	return *this;
}

//which simplifies (if unit vectors) to dot = cos(theta), the angle between them
inline float DotProd(const cpoint3& a, const cpoint3& b)
{
	return(a.x*b.x+a.y*b.y+a.z*b.z);
}




//------------------------------------------------------------------------------------------
// Point4
//------------------------------------------------------------------------------------------
class COREAPI cpoint4
{
	friend COREAPI cpoint4 MakePt4(float X, float Y, float Z, float W);

public:
	float x,y,z,w;

	cpoint4(){ x = 0.0f; y = 0.0f; z = 0.0f; w=0.0f;}
	cpoint4(const cpoint4& pt) { x = pt.x; y = pt.y; z = pt.z; w = pt.w; } 
	cpoint4(float X, float Y, float Z, float W)  { x = X; y = Y; z = Z; w = W; }
	cpoint4(int X, int Y, int Z, int W) { x = (float)X; y = (float)Y; z = (float)Z; w = (float)W; }
	cpoint4(float pt[4]) { x = pt[0]; y = pt[1]; z = pt[2]; w = pt[3]; }
	cpoint4(const cpoint3& pt) { x = pt.x; y = pt.y; z = pt.z; w = 0.0f; } 
	cpoint4(float X, float Y, float Z)  { x = X; y = Y; z = Z; w = 0.0f; }
	cpoint4(int X, int Y, int Z) { x = (float)X; y = (float)Y; z = (float)Z; w = 0.0f; }

	~cpoint4(){}//(ver.>=1.3.4):needed for Rayscript

	inline float Length() const;
	void Normalize();//in place

	//(ver.>=1.3.6):Bounding Box
	inline void AddMins(const cpoint4 &p);
	inline void AddMaxs(const cpoint4 &p);

	// Access operators
	float&  operator[](int i) { return (&x)[i]; }
	const float& operator[](int i) const { return (&x)[i]; }

	// Conversion function
	operator float*() { return (float*)(&x); }

	// Unary operators
	cpoint4 operator-() const { return(MakePt4(-x,-y,-z,-w)); } 
	cpoint4 operator+() const { return *this; }

	// Assignment operators
	inline cpoint4& operator-=(const cpoint4&);
	inline cpoint4& operator+=(const cpoint4&);
	inline cpoint4& operator/=(const cpoint4&);
	inline cpoint4& operator*=(const cpoint4&);
	inline cpoint4& operator-=(float);
	inline cpoint4& operator+=(float);
	inline cpoint4& operator/=(float);
	inline cpoint4& operator*=(float); 

	// Binary operators
	inline  cpoint4 operator-(const cpoint4&) const;
	inline  cpoint4 operator+(const cpoint4&) const;
	inline  cpoint4 operator/(const cpoint4&) const;
	inline  cpoint4 operator*(const cpoint4&) const;
	inline cpoint4 operator-(float) const;
	inline cpoint4 operator+(float) const;
	inline cpoint4 operator/(float) const;
	inline cpoint4 operator*(float) const;

	cpoint4 operator*(const cmatrix4& M) const;//(ver.>=1.3.4):Moved here.

	// Comparison operators
	inline int operator==(const cpoint4& p) const { return ((p.x==x)&&(p.y==y)&&(p.z==z)&&(p.w==w)); }
	inline int operator!=(const cpoint4& p) const { return ((p.x!=x)||(p.y!=y)||(p.z!=z)||(p.w!=w)); }

	//Max-Min component
	inline float Max() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z), fW=fabs(w);
		if(fX>=fY)
		{
			if(fX>=fZ)
			{
				if(fX>=fW) return x;
				return w;
			}
			if(fZ>=fW) return z;
			return w;
		}
		if(fY>=fZ)
		{
			if(fY>=fW) return y;
			return w;
		}
		if(fZ>=fW) return z;
		return w;
	}
	inline float Min() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z), fW=fabs(w);
		if(fX<=fY)
		{
			if(fX<=fZ)
			{
				if(fX<=fW) return x;
				return w;
			}
			if(fZ<=fW) return z;
			return w;
		}
		if(fY<=fZ)
		{
			if(fY<=fW) return y;
			return w;
		}
		if(fZ<=fW) return z;
		return w;
	}

	//Max-Min axis
	inline int MaxAxis() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z);
		if(fX>=fY)
		{
			if(fX>=fZ) return AXIS_X;
			return AXIS_Z;
		}
		if(fY>=fZ) return AXIS_Y;
		return AXIS_Z;
	}
	inline int MinAxis() const
	{
		float fX=fabs(x), fY=fabs(y), fZ=fabs(z);
		if(fX<=fY)
		{
			if(fX<=fZ) return AXIS_X;
			return AXIS_Z;
		}
		if(fY<=fZ) return AXIS_Y;
		return AXIS_Z;
	}

	//Interpolation
	inline cpoint4& Inter(const cpoint4& b, float t);
};

COREAPI cpoint4 Normalize(const cpoint4&);
COREAPI cpoint4 CrossProd(const cpoint4& a, const cpoint4& b);
COREAPI BOOL Compare(const cpoint4& a, const cpoint4& b,float tolerance);//tolerance >=0
COREAPI cpoint4 Inter(const cpoint4& a, const cpoint4& b, float t);



inline float cpoint4::Length() const
{
	return (float)sqrt(x*x+y*y+z*z);
}

//(ver.>=1.3.6):Bounding Box
inline void cpoint4::AddMins(const cpoint4 &p)
{
	if(p.x < x) x = p.x;
	if(p.y < y) y = p.y;
	if(p.z < z) z = p.z;
	if(p.w < w) w = p.w;
}

//(ver.>=1.3.6):Bounding Box
inline void cpoint4::AddMaxs(const cpoint4 &p)
{
	if(p.x > x) x = p.x;
	if(p.y > y) y = p.y;
	if(p.z > z) z = p.z;
	if(p.w > w) w = p.w;
}

inline cpoint4& cpoint4::operator-=(const cpoint4& a)
{	
	x -= a.x;	y -= a.y;	z -= a.z;   w -= a.w;
	return *this;
}

inline cpoint4& cpoint4::operator+=(const cpoint4& a)
{
	x += a.x;	y += a.y;	z += a.z;   w += a.w;
	return *this;
}

inline cpoint4& cpoint4::operator/=(const cpoint4& a)
{ 
	x /= a.x;	y /= a.y;	z /= a.z;   w /= a.w;
	return *this; 
}

inline cpoint4& cpoint4::operator*=(const cpoint4& a)
{ 
	x *= a.x;	y *= a.y;	z *= a.z;	w *= a.w;
	return *this; 
}

inline cpoint4& cpoint4::operator-=(float f)
{
	x -= f;   y -= f;	z -= f;   w -= f;
	return *this;
}

inline cpoint4& cpoint4::operator+=(float f)
{
	x += f;   y += f;	z += f;   w += f;
	return *this;
}

inline cpoint4& cpoint4::operator/=(float f)
{
	x /= f;   y /= f;	z /= f;	  w /= f;
	return *this;
}

inline cpoint4& cpoint4::operator*=(float f)
{
	x *= f;   y *= f;	z *= f;   w *= f;
	return *this;
}

inline cpoint4 cpoint4::operator-(const cpoint4& b) const
{
	return(MakePt4(x-b.x,y-b.y,z-b.z,w-b.w));
}

inline cpoint4 cpoint4::operator+(const cpoint4& b) const
{
	return(MakePt4(x+b.x,y+b.y,z+b.z,w+b.w));
}

inline cpoint4 cpoint4::operator/(const cpoint4& b) const
{
	return MakePt4(x/b.x,y/b.y,z/b.z,w/b.w);
}

inline cpoint4 cpoint4::operator*(const cpoint4& b) const
{
	return MakePt4(x*b.x, y*b.y, z*b.z , w*b.w);	
}

inline cpoint4 cpoint4::operator-(float f) const
{
	return(MakePt4(x-f, y-f, z-f , w-f));
}

inline cpoint4 cpoint4::operator+(float f) const
{
	return(MakePt4(x+f, y+f, z+f , w+f));
}

inline cpoint4 cpoint4::operator/(float f) const
{
	return(MakePt4(x/f, y/f, z/f , w/f));
}

inline cpoint4 cpoint4::operator*(float f) const
{
	return(MakePt4(x*f, y*f, z*f , w*f));
}

inline cpoint4& cpoint4::Inter(const cpoint4& b,float t)
{	
	x = INTERPOLATE(x, b.x, t);
	y = INTERPOLATE(y, b.y, t);
	z = INTERPOLATE(z, b.z, t);
	w = INTERPOLATE(w, b.w, t);
	return *this;
}

//which simplifies (if unit vectors) to dot = cos(theta), the angle between them
inline float DotProd(const cpoint4& a, const cpoint4& b)
{
	return(a.x*b.x+a.y*b.y+a.z*b.z);
}




//------------------------------------------------------------------------------------------
// Matrix4
//------------------------------------------------------------------------------------------
typedef float matrix4[4][4];

class COREAPI cmatrix4
{
	matrix4 m;
public:

	// Constructors
	cmatrix4();//default constructor = identity matrix
	cmatrix4(matrix4& M);

	~cmatrix4(){}//(ver.>=1.3.4):needed for Rayscript

    cmatrix4& Set(const matrix4& M);
    cmatrix4& Set(const cpoint3& X, const cpoint3& Y, const cpoint3& Z, const cpoint3& T)
	{
		SetRow(0, X); SetRow(1, Y); SetRow(2, Z); SetRow(3, T);
		return *this;
	}
    cmatrix4& Set(const cpoint4& X, const cpoint4& Y, const cpoint4& Z, const cpoint4& T)
	{
		SetRow(0, X); SetRow(1, Y); SetRow(2, Z); SetRow(3, T);
		return *this;
	}

	void GetM(matrix4& M) const;

	void Identity(); 		// Make into the Identity Matrix
	void Zero();			// Set all elements to 0

	BOOL IsIdentity() const;
	BOOL IsOrthogonal() const;	// (row1 cross row2 = row3 & col1 cross col2 = col3)
	BOOL IsOrthonormal() const;	// (if the rows and columns are all normalized (transform has no scaling or shearing)
								// and is orthogonal (row1 cross row2 = row3 & col1 cross col2 = col3)

	void SetRow(int i, cpoint3 p) { (*this)[i][0] = p.x; (*this)[i][1] = p.y; (*this)[i][2] = p.z; }
	void SetRow(int i, float x,float y,float z) { (*this)[i][0] = x; (*this)[i][1] = y; (*this)[i][2] = z; }
	cpoint3 GetRow3(int i) const { return MakePt3((*this)[i][0],(*this)[i][1],(*this)[i][2]); }
	void SetRow(int i, cpoint4 p) { (*this)[i] = p; }
	void SetRow(int i, float x,float y,float z,float w) { (*this)[i][0] = x; (*this)[i][1] = y; (*this)[i][2] = z; (*this)[i][3] = w; }
	cpoint4 GetRow4(int i) const { return (*this)[i]; }

	void SetTrans(const cpoint3 p) { (*this)[3][0] = p.x; (*this)[3][1] = p.y; (*this)[3][2] = p.z; }
	void SetTrans(float x,float y,float z) { (*this)[3][0] = x; (*this)[3][1] = y; (*this)[3][2] = z; }
	void SetTrans(int i, float v) { (*this)[3][i] = v; }
	cpoint3 GetTrans() const { return MakePt3((*this)[3][0],(*this)[3][1],(*this)[3][2]); }

	void Orthonormalize();//removes scaling (or other distortions) from an orthogonal (or nearly orthogonal) matrix 
    void Invert();// in place
	void Transpose();//in place
	void RemoveScaling();//(ver.>=1.3.6):in place - normalizes x,y,z rows
	cpoint3 GetScale() const;//(ver.>=1.3.6):returns magnitudes of x,y,z rows as x,y,z scale values

	//matrix from translation
	void FromTranslation(float ix,float iy,float iz);
	void FromTranslation(const cpoint3&);
	//matrix from euler angles
	void FromAngleX(float angle);
	void FromAngleY(float angle);
	void FromAngleZ(float angle);
	// order of angles z,y,x
	void FromAngles(float x,float y,float z);
	void FromAngles(const cpoint3&);
	//matrix from scale
	void FromScale(float sx,float sy,float sz);
	void FromScale(const cpoint3&);
	//matrix from shear
	void FromShear(float thetax,float thetay,float thetaz);
	void FromShear(const cpoint3&);

	// Apply Incremental transformations to this matrix
	void Translate(const cpoint3& p);
	void RotateX(float angle);
	void RotateY(float angle);
	void RotateZ(float angle);
	// if trans = TRUE the translation component will be affected.
	void Scale(const cpoint3& s, BOOL trans = FALSE);

	//Transform point by this matrix
	cpoint3 PointTransform(const cpoint3& p) const;
	cpoint4 PointTransform(const cpoint4& p) const;
	cpoint3 operator*(const cpoint3& p) const;//(ver.>=1.3.4):Moved here.
	cpoint4 operator*(const cpoint4& p) const;//(ver.>=1.3.4):Moved here.

	// Access operators
//	cpoint3& operator[](int i) { return((cpoint3&)(*m[i]));  }
//	const cpoint3& operator[](int i) const { return((cpoint3&)(*m[i])); }
	cpoint4& operator[](int i) { return((cpoint4&)(*m[i]));  }
	const cpoint4& operator[](int i) const { return((cpoint4&)(*m[i])); }

	// Conversion function
	operator float*() { return (float*)m; }

    // Comparison operator
//	int operator==(const cmatrix4&) const;
	// Assignment operators
//	cmatrix4& operator-=(const cmatrix4&);
//	cmatrix4& operator+=(const cmatrix4&); 
	cmatrix4& operator*=(const cmatrix4&);
	//Binary operators
//	cmatrix4 operator-(const cmatrix4&) const;
//	cmatrix4 operator+(const cmatrix4&) const;
	cmatrix4 operator*(const cmatrix4&) const;
};

// Returns inverse of matrix
COREAPI cmatrix4 Inverse(const cmatrix4& M);
// The transpose is rapidly computed and is equal to the inverse 
// transform for orthonormal matrices
COREAPI cmatrix4 Transpose(const cmatrix4& M);
COREAPI cmatrix4 RemoveScaling(const cmatrix4& M);//(ver.>=1.3.6):normalizes x,y,z rows
COREAPI cpoint3 GetScale(const cmatrix4& M);//(ver.>=1.3.6):returns magnitudes of x,y,z rows as x,y,z scale values


COREAPI cpoint3 EulerFromMatrix(const cmatrix4& M);
COREAPI void EulerFromMatrix(const cmatrix4& M,float ang[3]);
// order of angles z,y,x
COREAPI void MatrixFromEuler(cmatrix4& M, const cpoint3& ang);
COREAPI void MatrixFromEuler(cmatrix4& M, const float ang[3]);

// Make rotation matrix out of dir,up vectors.Won't affect translation vector
COREAPI void MatrixLookAt(cmatrix4& M,const cpoint3& dirz,const cpoint3& up);


//------------------------------------------------------------------------------------------
// Quaternion
//------------------------------------------------------------------------------------------
class COREAPI cquat
{
	friend COREAPI cquat MakeQuat(float X, float Y, float Z, float W);

public:
	float x,y,z,w;

	cquat() { x=0.0f;y=0.0f;z=0.0f;w=1.0f; }
	cquat(float X, float Y, float Z, float W) { x = X; y = Y; z = Z; w = W; }
	cquat(const cquat& q) { x = q.x; y = q.y; z = q.z; w = q.w; } 
	cquat(float q[4]) { x = q[0]; y = q[1]; z = q[2]; w = q[3]; }
	cquat(const cpoint3& Axis, float Angle);
	cquat(const cpoint4& p) { x = p.x; y = p.y; z = p.z; w = p.w; } 
	cquat(const cmatrix4& m);
	cquat(const cpoint3& euler);
	cquat(float eulerX,float eulerY,float eulerZ);

	~cquat(){}//(ver.>=1.3.4):needed for Rayscript

	cquat& Set(cpoint4 p) { x = p.x; y = p.y; z = p.z; w = p.w; return *this; }
	cquat& Set(cpoint3 p,float W) { x = p.x; y = p.y; z = p.z; w = W; return *this; }
	cquat& Set(float X, float Y, float Z, float W) { x = X; y = Y; z = Z; w = W; return *this; }
	cquat& FromAngleAxis(const cpoint3& Axis, float Angle); 
	cquat& FromMatrix(const cmatrix4& mat);
	// order of angles z,y,x
	cquat& FromEuler(const cpoint3& euler);
	cquat& FromEuler(float X, float Y, float Z);

	void Invert();// in place

	void Identity() { x=0.0f;y=0.0f;z=0.0f;w=1.0f; }
	BOOL IsIdentity() const;
	BOOL IsUnit() const;
	void Normalize();  // in place
	cpoint3 Vector() const { return MakePt3(x, y, z); }
	float Length() const;
	cquat& Conjugate();//(ver.>=1.3.6):added
	//Rotate point by this quaternion
	cpoint3 PointRotate(const cpoint3& p) const;

	// Access operators
	float& operator[](int i) { return (&x)[i]; }     
	const float& operator[](int i) const { return (&x)[i]; }
    
	// Conversion function
	operator float*() { return (float*)(&x); }

	// Unary operators
	cquat operator-() const { return(MakeQuat(-x,-y,-z,w/*-w*/)); }//conjugate
	cquat operator+() const { return *this; }

	// Assignment operators
	cquat& operator-=(const cquat&);
	cquat& operator+=(const cquat&);
	cquat& operator*=(const cquat&);
	cquat& operator/=(const cquat&);//(ver.>=1.3.5 update1):Division means this*Inverse(q)
	cquat& operator*=(float);
	cquat& operator/=(float);

	// Comparison operator
	int operator==(const cquat& q) const { return ((q.x==x)&&(q.y==y)&&(q.z==z)&&(q.w==w)); }
	int operator!=(const cquat& q) const { return ((q.x!=x)||(q.y!=y)||(q.z!=z)||(q.w!=w)); }

	// Binary operators
	cquat operator-(const cquat&) const;  // difference of two quaternions
	cquat operator+(const cquat&) const;  // sum of two quaternions
	cquat operator*(const cquat&) const;  // product of two quaternions
	cquat operator/(const cquat&) const;  // ratio of two quaternions
										  // (ver.>=1.3.5 update1):Division means this*Inverse(q)
	cquat operator*(float) const;		  // multiply by scalar
	cquat operator/(float) const;		  // divide by scalar
};

COREAPI cquat Normalize(const cquat&);
COREAPI cquat Inverse(const cquat&);// Inverse of quaternion (1/q)
COREAPI cquat Conjugate(const cquat& q);//(ver.>=1.3.6):added
COREAPI cquat Slerp(const cquat& q1, const cquat& q2, float t);
COREAPI cquat QuatFromAngleAxis(float ang, const cpoint3& axis);
COREAPI cquat QuatFromMatrix(const cmatrix4& m);
COREAPI void MatrixFromQuat(const cquat& q,cmatrix4& m);
COREAPI cquat TransformQuat(const cmatrix4& m, const cquat& q);
COREAPI cmatrix4 TransformMatrix(const cmatrix4& m, const cquat& q);

COREAPI void QuatToEuler(const cquat& q, float ang[3]);
COREAPI cpoint3 QuatToEuler(const cquat& q);
// order of angles x,y,z
COREAPI cquat EulerToQuat(const float ang[3]);
COREAPI cquat EulerToQuat(const cpoint3& ang);
// Transform Point by Quaternion
COREAPI cpoint3 operator*(const cquat& Q, const cpoint3& V);
COREAPI cpoint3 operator*(const cpoint3& V,const cquat& Q);



//------------------------------------------------------------------------------------------
// AngleAxis
//------------------------------------------------------------------------------------------
class COREAPI cangleaxis
{
	friend COREAPI cangleaxis MakeAngleAxis(float X, float Y, float Z, float Angle);
	friend COREAPI cangleaxis MakeAngleAxis(const cpoint3& Axis, float Angle);

public:
	cpoint3 axis;
	float angle;

	cangleaxis() { axis=MakePt3(0.0f,0.0f,0.0f);angle=0.0f; }
	cangleaxis(const cangleaxis& a) { *this=a; } 
	cangleaxis(float X, float Y, float Z, float Angle) { axis.x = X; axis.y = Y; axis.z = Z; angle = Angle; }
	cangleaxis(const cpoint3& Axis, float Angle) { axis = Axis; angle = Angle; }
	cangleaxis(float q[4]) { axis.x = q[0]; axis.y = q[1]; axis.z = q[2]; angle = q[3]; }
	cangleaxis(const cquat& q); 
	cangleaxis(const cmatrix4& m);
	cangleaxis(const cpoint3& euler);
	cangleaxis(float eulerX,float eulerY,float eulerZ);

	~cangleaxis(){}//(ver.>=1.3.4):needed for Rayscript

	cangleaxis& Set(cpoint3& Axis, float Angle) { axis = Axis; angle = Angle; return *this; }
	cangleaxis& Set(float X, float Y, float Z, float Angle) { axis.x = X; axis.y = Y; axis.z = Z; angle = Angle; return *this; }
	cangleaxis& FromQuat(const cquat&); 
	cangleaxis& FromMatrix(const cmatrix4&);
	// order of angles z,y,x
	cangleaxis& FromEuler(const cpoint3& euler);
	cangleaxis& FromEuler(float X, float Y, float Z);

	void Invert();// in place

	void Identity() { axis=MakePt3(0.0f,0.0f,0.0f);angle=0.0f; }
	BOOL IsIdentity() const;

	// Unary operators
	cangleaxis operator-() const { return(MakeAngleAxis(-axis.x,-axis.y,-axis.z,angle/*-angle*/)); } 
	cangleaxis operator+() const { return *this; }

	/*
	// Assignment operators
	cangleaxis& operator-=(const cangleaxis&);
	cangleaxis& operator+=(const cangleaxis&);
	cangleaxis& operator*=(const cangleaxis&);
	cangleaxis& operator/=(const cangleaxis&);
	cangleaxis& operator*=(float);
	cangleaxis& operator/=(float);
	*/

	// Comparison operator
	int operator==(const cangleaxis& a) const { return ((a.axis==axis)&&(a.angle==angle)); }
	int operator!=(const cangleaxis& a) const { return ((a.axis!=axis)||(a.angle!=angle)); }

	/*
	// Binary operators
	cangleaxis operator-(const cangleaxis&) const;
	cangleaxis operator+(const cangleaxis&) const;
	cangleaxis operator*(const cangleaxis&) const;
	cangleaxis operator/(const cangleaxis&) const;
	cangleaxis operator*(float) const;
	cangleaxis operator/(float) const;
	*/
};

COREAPI cangleaxis Inverse(const cangleaxis&);
COREAPI void AngleAxisFromQuat(const cquat& q, float *ang, cpoint3& axis);
COREAPI cangleaxis AngleAxisFromQuat(const cquat& q);



//------------------------------------------------------------------------------------------
// Helpers
//------------------------------------------------------------------------------------------

//------------------------------------------------------------------------------------------
// 2D helper funcs

//Check if point is in poly.
//cpoint2 version
COREAPI BOOL PtInPoly(cpoint2 *points,int *poly,int numV,float x,float y);
COREAPI BOOL PtInPoly(cpoint2 *points,int numV,float x,float y);//Provided points order describes a poly
//cpoint3 version(z component ignored)
COREAPI BOOL PtInPoly(cpoint3 *points,int *poly,int numV,float x,float y);
COREAPI BOOL PtInPoly(cpoint3 *points,int numV,float x,float y);//Provided points order describes a poly

//Check if the point (x,y) is near the line (x1,y1-x2,y2)
COREAPI BOOL NearPoint(float x1,float y1,float x2,float y2,float x,float y,float delta);

//Find intersection of lines defined by (a1,a2) and (b1,b2).(z component is ignored)
//Returns 1 if valid intersection, 0 if parallel lines
//If bBothMustIncludePoint is set then returns 1 only if not parallel lines and both lines include the point
COREAPI int LinesIntersection(cpoint3 a1, cpoint3 a2, cpoint3 b1, cpoint3 b2, cpoint3 &p, BOOL bBothMustIncludePoint=FALSE);
//cpoint2 version
COREAPI int LinesIntersection(cpoint2 a1, cpoint2 a2, cpoint2 b1, cpoint2 b2, cpoint2 &p, BOOL bBothMustIncludePoint=FALSE);


//------------------------------------------------------------------------------------------
//(ver.>=1.3.5 update1):Added 3D helper funcs

// Get the shortest distance from point to line.
//    Input:  a point P and a Line LP0,LP1 (infinite)
//    (z=0 for 2D)
COREAPI float PointToLineDist(cpoint3 P, cpoint3 LP0, cpoint3 LP1);
// Get the shortest distance from point to segment.
//    Input:  a point P and a Segment SP0,SP1
//    (z=0 for 2D)
COREAPI float PointToSegmentDist(cpoint3 P, cpoint3 SP0, cpoint3 SP1);
// Get the shortest distance between 2 lines.
//    Input:  two 3D lines L1 and L2 (infinite)
COREAPI float LineToLineDist( cpoint3 L1P0, cpoint3 L1P1, cpoint3 L2P0, cpoint3 L2P1);
// Get the shortest distance between 2 segments.
//    Input:  two 3D line segments S1 and S2
COREAPI float SegmentToSegmentDist( cpoint3 S1P0, cpoint3 S1P1, cpoint3 S2P0, cpoint3 S2P1);


//------------------------------------------------------------------------------------------
//Plane helpers

COREAPI void PlaneEquation(cpoint3 A,cpoint3 B,cpoint3 C,cpoint3& n,float *D);
COREAPI float FindPlaneD(cpoint3 A,cpoint3 B,cpoint3 C);
inline COREAPI float PlaneD(cpoint3 normal,cpoint3 v){return (-DotProd(normal,v));}
COREAPI float DistanceToPlane(float x,float y,cpoint3 normal,float D);
COREAPI float DistanceToPlane(float x,float y,cpoint3 A,cpoint3 B,cpoint3 C);
inline COREAPI float DistanceToPlane(cpoint3 p, cpoint3 pnormal,float D){return(DotProd(pnormal,p) + D);}
// 3 points a,b,c define a plane: compare point p with them.
COREAPI BOOL PointInsideFace(cpoint3 A, cpoint3 B, cpoint3 C, cpoint3 p, float *dist);


//------------------------------------------------------------------------------------------
// Catmull-Rom Curve
COREAPI cpoint3 CatmullRom(cpoint3 *p4, float t);//*p[4]


//------------------------------------------------------------------------------------------
// Test for concavity(points must be contiguous)
COREAPI BOOL IsPolyConcave(cpoint3 *points,int npoints);//TRUE=concave,FALSE=convex


//------------------------------------------------------------------------------------------
//(ver.>=1.3.6 update1): Snap to grid

inline COREAPI void SnapVector(cpoint2& v)
{
	v.x=SNAPFLOAT(v.x);
	v.y=SNAPFLOAT(v.y);
}

inline COREAPI void SnapVectorToGrid(cpoint2& v,float gridsize)
{
	v.x = SNAPTOGRID(v.x,gridsize);
	v.y = SNAPTOGRID(v.y,gridsize);
}

inline COREAPI void SnapVector(cpoint3& v)
{
	v.x=SNAPFLOAT(v.x);
	v.y=SNAPFLOAT(v.y);
	v.z=SNAPFLOAT(v.z);
}

inline COREAPI void SnapVectorToGrid(cpoint3& v,float gridsize)
{
	v.x = SNAPTOGRID(v.x,gridsize);
	v.y = SNAPTOGRID(v.y,gridsize);
	v.z = SNAPTOGRID(v.z,gridsize);
}

inline COREAPI void SnapVector(cpoint4& v)
{
	v.x=SNAPFLOAT(v.x);
	v.y=SNAPFLOAT(v.y);
	v.z=SNAPFLOAT(v.z);
	v.w=SNAPFLOAT(v.w);
}

inline COREAPI void SnapVectorToGrid(cpoint4& v,float gridsize)
{
	v.x = SNAPTOGRID(v.x,gridsize);
	v.y = SNAPTOGRID(v.y,gridsize);
	v.z = SNAPTOGRID(v.z,gridsize);
	v.w = SNAPTOGRID(v.w,gridsize);
}



#endif //CMATHLIB_H