// Lightray3D (C) 2002-2006 Tzivenis Sakis

#ifndef CNODE_H
#define CNODE_H

#include "stdtypes.h"
#include "ccolor.h"
#include "cmathlib.h"
#include "canim.h"
#include "cmaterial.h"
#include "glwnd.h"
#include "cio.h"
#include "cbitarray.h"
#include "rendermesh.h"


//(ver.>=1.3.5):Render flags passed to Render(int attime,CGLWindow* view,UINT rflags)
#define RENDER_SOLID		1
#define RENDER_TRANS		2

//(ver.>=1.3.6):Increased to 64
#define MAX_NODE_NAME		64//32

//(ver.>=1.3.1):Changed MAX_NOTES_SIZE from 255 to 8192
#ifndef MAX_NOTES_SIZE
#define MAX_NOTES_SIZE		8192//255
#endif

//(ver.>=1.3.2):Added message notification system
#define MSG_NO_MSG					0
#define MSG_NODE_DELETED			1//Param1 is a cnode*
#define MSG_MAT_REMOVED				2//Param1 is a cmaterial*
//(ver.>=1.3.3)
#define MSG_NODE_CLONED				3//Param1 is original node,Param2 is new cloned node
//NOTE:Check derived classes *.h files as they may have their own MSG defined
#define MSG_USER					10000//future use

//Selection flags
#define SEL_ISHIT		-1	//Just do a hit test, no selecting.Return index
#define SEL_SEL			0	//Select what's hit, deselect others
#define SEL_ADD			1	//Select what's hit, preserve others' selection state
#define SEL_SUB			2	//Deselect what's hit, preserve others' selection state

//State Flags(valid for elements too)
#define NO_FLAGS			0
#define FLAG_SELECTED		(1<<0)
#define FLAG_HIDDEN			(1<<1)
#define FLAG_HOLD			(1<<2)//Removed from scene(kinda deleted) but still in memory until undo restores it.
								  //Objects using an object pointer should not use it if this flag is set.

//Object level
#define SUB_LEVEL_OBJECT		0
#define SUB_LEVEL_POINT			1
#define SUB_LEVEL_POLYGON		2
#define SUB_LEVEL_LINE			3

//Element type
#define ELEMENT_UNKNOWN			0
#define ELEMENT_POINT			1
#define ELEMENT_POLYGON			2
#define ELEMENT_LINE			3
#define ELEMENT_TVPOINT			4

class COREAPI celement
{
public:
	DWORD type;		//Element type
	WORD wFlags;	//State Flags

	celement(){type=ELEMENT_UNKNOWN;wFlags=NO_FLAGS;}
	~celement(){}//(ver.>=1.3.4):Needed by RAYscript
};

//Node transform parsing options
#define TM_TRANS	(1<<0)
#define TM_ROT		(1<<1)
#define TM_SCALE	(1<<2)
#define TM_ALL		(TM_SCALE|TM_ROT|TM_TRANS)

//Default hit delta value
#define HITDELTA	2.0f

//IK
#define IK_DEFAULT_DAMP			10.0f
#define IK_DEFAULT_MIN_ANGLE	-180.0f
#define IK_DEFAULT_MAX_ANGLE	180.0f
#define IK_DEFAULT_ITERATIONS	100
#define IK_MAX_ITERATIONS		200
#define IK_DEFAULT_RECURSE_HIERARCHY	2//up to parent's parent bone
#define IK_MAX_RECURSE_HIERARCHY		1000//up to 1000 nodes in hierarchy(nested enough???)

class COREAPI cnode
{
	UINT type,subtype;
	char name[MAX_NODE_NAME];
public:
	cnode();
	cnode(const char *Name);
	cnode(const char *Name,const ccolor& clr);
	virtual ~cnode();//will take care of informing parent,children
	
	virtual void Delete();

	virtual void SetName(const char*);
	virtual void GetName(char*) const;
	virtual char* GetName();

	//Class type
	virtual UINT GetTypeID() const {return type;}
	virtual UINT GetSubTypeID() const {return subtype;}

	//Flags
	virtual void SetFlags(DWORD);
	virtual DWORD GetFlags() const;

	//Sublevel
	virtual void SetSubLevel(int);
	virtual int GetSubLevel();

	//Wireframe color
	virtual void SetColor(const ccolor& clr){color=clr;}
	virtual ccolor GetColor() const {return color;}

	//Bounding Box(LocalCS).Return FALSE if no bbox
	virtual BOOL GetBBox(int attime,cpoint3& bmin,cpoint3& bmax) {bmin=min;bmax=max;return TRUE;}

	//UV mapping
	virtual BOOL HandlesUVMapping() const {return FALSE;}
	//type: 0=plane 1=box 2=cylinder 3=sphere
	//(ver.>=1.3.5:)Added capping flag(cylindrical mapping)
	virtual BOOL ApplyUV(int type,BOOL bCap,BOOL bFlipU,BOOL bFlipV,float uscale,float vscale,const cmatrix4& tm)
	{
		return FALSE;
	}

	//Render(Geometry objects should handle these)
	virtual BOOL IsRenderable() const {return FALSE;}
	virtual RenderMesh* GetRenderMesh(int attime) {return NULL;}

	//Hierarchy
	//When linking to parent all coordinates(animation keys and other) are transformed
	//to parent space.Set bConvToParent to FALSE if you want them unchanged
	virtual void SetParent(const cnode*parent,BOOL bConvToParent=TRUE);
	virtual cnode* GetParent() const;
	//If bSetParent==FALSE you must call child's SetParent to notify it
	virtual void Attach(cnode*child,BOOL bSetParent=TRUE);
	virtual void Detach(const cnode*child,BOOL bSetParent=TRUE);
	virtual void Detach(int at,BOOL bSetParent=TRUE);
	virtual void DetachAll(BOOL bSetParent=TRUE);
	virtual cnode* GetChild(int at) const;
	virtual int GetChild(const cnode*child) const;
	virtual int NumChildren() const {return nchildren;}

	//Material(TODO:change pointers to index maybe?)
	virtual void SetMaterial(cmaterial*){}
	virtual cmaterial* GetMaterial() const{return NULL;}

	//Clone
	virtual BOOL Clone(int attime,cnode&destnode) const;

	//World space coordinates(animatable)
	//NOTE:If attime==TIME_IDLE won't set key(you must set key directly to anim,see GetAnimPtr)
	//unless record is on else will set key even if record is off
	virtual void SetPosition(int attime,const cpoint3&);
	virtual void SetRotation(int attime,const cquat&);
	virtual void SetScale(int attime,const cpoint3&);
	virtual void SetNodeTM(int attime,const cmatrix4&,int tmflags=TM_ALL);//(ver.>=1.3.6)
	virtual cpoint3 GetPosition(int attime) const;
	virtual cquat GetRotation(int attime) const;
	virtual cpoint3 GetScale(int attime) const;
	virtual cmatrix4 GetNodeTM(int attime,int tmflags=TM_ALL) const;

	//Animation(Note:If parent exists keys hold parent space values else world space)
	canim* GetAnimPtr(){return &anim;}
	//Returns TRUE if key exists
	virtual BOOL HasKey(int attime);
	//Called every time animation keys change or if animation current time changes
	virtual void UpdateTimeIdle();
	//Snap key at time idle(sets all animatable values at 'current time')
	virtual void SnapKey();
	//Delete key 'at time'
	virtual void DeleteKey(int attime);
	//Move keys(for single key pos==from)
//	virtual void MoveKeys(int pos,int from,int to);//(ver.>=1.3.1):Let app handle this

	//Display
	virtual void Flash(int attime,CGLWindow*);//if not overriden will flash bbox
	//(ver.>=1.3.5):Added render flags parameter
	virtual void Render(int attime,CGLWindow*,UINT rflags){}//Make this pure virtual?

	//Rendering(geometry specific)
	virtual void SetCastShadowFlag(BOOL);
	virtual void SetReceiveShadowFlag(BOOL);
	virtual BOOL GetCastShadowFlag() const;
	virtual BOOL GetReceiveShadowFlag() const;

	//Elements
	//(ver.>=1.3.6):Added GetElement()
	virtual BOOL GetElement(celement& elem,int index,UINT elemtype){return FALSE;}//Returns FALSE if index out of range OR not such element type
	virtual int GetNumElements(UINT elemtype){return 0;}
	virtual int GetNumSelElements(UINT elemtype){return 0;}
	virtual int GetNumHiddenElements(UINT elemtype){return 0;}//(ver.>=1.3.1)
	virtual BOOL GetSelElementsCenter(UINT elemtype,int attime,cpoint3& c)//LCS
	{
	//	c=MakePt3(0,0,0);
		return FALSE;
	}
	//Sub-level specific
	virtual int SelectAll(){return 0;}
	virtual int DeselectAll(){return 0;}
	virtual int InvertSelection(){return 0;}
	virtual void HideSelected(){}
	virtual void UnhideAll(){}
	virtual void DeleteSelected(){}
	//Returns hit element's index
	virtual int PickElement(int attime,CGLWindow*,int x,int y,int selflags=SEL_SEL,float delta=HITDELTA){return -1;}
	//These return (min z value)hit element's index if selflags==SEL_ISHIT or number of hit elements
	virtual int PickElementRect(int attime,CGLWindow*,int x,int y,int x2,int y2,int selflags=SEL_SEL){return 0;}
	virtual int PickElementCircle(int attime,CGLWindow*,int x,int y,int radius,int selflags=SEL_SEL){return 0;}
	virtual int PickElementContour(int attime,CGLWindow*,cpoint3*,int,int selflags=SEL_SEL){return 0;}
	//Usually called after mouse drags.
	//(ver.>=1.3.2):Added pivot flag.Only derived classes that own geometry should handle this flag(e.g mesh,shape)
	//(ver.>=1.3.6):Added bUseViewIKThresh,ikthresh
	virtual void Move(int attime,const cpoint3&,BOOL bPivotOnly=FALSE,BOOL bIK=FALSE,BOOL bUseViewIKThresh=TRUE,float ikthresh=1.0f);//WCS
	virtual void Rotate(int attime,const cquat&,const cpoint3& center,BOOL bPivotOnly=FALSE);//WCS
	virtual void Scale(int attime,const cpoint3&,const cpoint3& center,BOOL bPivotOnly=FALSE);//LCS

	//(ver.>=1.3.5):Added virtual Mirror func.
	//If tm is passed, transform geometry before mirroring.
	virtual BOOL Mirror(int attime,int axis,cmatrix4* tm=NULL){return FALSE;}//FALSE = not handled or error

	//Picking.Checks hits on node center(pivot) position
	virtual BOOL HitTest(int attime,CGLWindow*,int x,int y,float *zvalue,float delta=HITDELTA);
	virtual BOOL HitTestRect(int attime,CGLWindow*,int x,int y,int x2,int y2);
	virtual BOOL HitTestCircle(int attime,CGLWindow*,int x,int y,int radius);
	virtual BOOL HitTestContour(int attime,CGLWindow*,cpoint3*,int);
	//Check hits inside node's BBox
	//Can be called from inside descendants HitTest overriden funcs.
	BOOL HitTestBBox(int attime,CGLWindow*,int x,int y,float *zvalue,float delta=HITDELTA);
	BOOL HitTestRectBBox(int attime,CGLWindow*,int x,int y,int x2,int y2);
	BOOL HitTestCircleBBox(int attime,CGLWindow*,int x,int y,int radius);
	BOOL HitTestContourBBox(int attime,CGLWindow*,cpoint3*,int);

	//(ver.>=1.3.4):IK
	void ResetIK();//set defaults
	//Degrees Of Freedom
	void EnableIKDOF(BOOL);
	BOOL IsIKDOFEnabled() const;
	void SetIKMinRot(const cpoint3&);
	cpoint3 GetIKMinRot() const;
	void SetIKMaxRot(const cpoint3&);
	cpoint3 GetIKMaxRot() const;
	void CalcIKInitialRot(int attime);//must call this once if DOF is gonna be used
	void SetIKInitialRot(const cquat&);//manually set(calculated) initial IK angles(rotation)
	cquat GetIKInitialRot() const;
	//Damping
	void EnableIKDamping(BOOL);
	BOOL IsIKDampingEnabled() const;
	void SetIKDampWidth(float);
	float GetIKDampWidth() const;
	//terminator
	void LockIK(BOOL b);
	BOOL IsIKLocked() const;
	//Other
	//Iterate until close to goal.
	void SetIKNumIterations(int);//Clamped between 1 to MAX_ITERATIONS.
	int GetIKNumIterations() const;
	//Recurse up the hierarchy(1 - up to parent, 2 - up to parent's parent etc.)
	void SetIKRecurseNum(int);//Clamped(between 1 to MAX_RECURSE)
	int GetIKRecurseNum() const;

	//(ver.>=1.3.6):See through
	virtual void SetSeeThroughFlag(BOOL bFlag);
	virtual BOOL GetSeeThroughFlag();

	//IO methods
	virtual BOOL Load(cio_load& iload,DWORD fileversion);//support for previous file versions
	virtual BOOL Save(cio_save& isave);

	//User defined notes(up to MAX_NOTES_SIZE, size+1 chr will-auto set to \0)
	virtual void ClearNotes();
	virtual void CopyNotes(const char* notes,int size);//will replace current notes array
	virtual void AddNotes(const char* notes,int size);//will append notes sizing it accordingly
	virtual char* GetNotes() const;//will return a temp pointer
	virtual int NotesSize() const;

	//(ver.>=1.3.2):Added single function for handling notification messages
	//like pointer variables of deleted nodes etc.
	//Must return TRUE if message handled succesfully,FALSE if not.
	virtual BOOL HandleMessage(DWORD msg,void *Param1,void *Param2);

	//These two are now used through messages:MSG_NODE_DELETED,MSG_MAT_REMOVED
	/*
	//Must handle pointer variables of deleted node.Called by undo manager on final deletion.
	virtual void NodeIsBeingDeleted(cnode*delnode){}

	//Must reset material and return TRUE if material is used by this node
	virtual BOOL MaterialIsBeingRemoved(cmaterial*mat)
	{
		if(GetMaterial()==mat)
		{
			SetMaterial(NULL);
			return TRUE;
		}
		return FALSE;
	}
	*/
	//Must return TRUE if material is used by this node
	virtual BOOL IsMaterialUsed(cmaterial*mat)
	{
		if(GetMaterial()==mat) return TRUE;
		return FALSE;
	}

	//(ver.>=1.3.2):Added pivot virtual functions
	//(ver.>=1.3.6):Used TIME_IDLE, now using specified frame('attime' param)
	virtual void ResetTransform(int attime,int tmflags=(TM_ROT|TM_SCALE));//(ver.>=1.3.6 update1):Added tmflags, choose rotation,scale or both
	virtual void CenterGeomToPivot(int attime);
	virtual void CenterPivotToGeom(int attime);
	void CenterPivotToGeom(int attime,const cpoint3& geomcenterinlcs);//Don't use.Called from child classes that override CenterPivotToGeom()


protected:
	void SetTypeID(UINT t){type=t;}
	void SetSubTypeID(UINT t){subtype=t;}

	void Remove(int);
	cpoint3 PosToParent(int,const cpoint3&) const;
	cquat RotToParent(int,const cquat&) const;
	cpoint3 ScaleToParent(int,const cpoint3&) const;
	cpoint3 PosFromParent(int,const cpoint3&) const;
	cquat RotFromParent(int,const cquat&) const;
	cpoint3 ScaleFromParent(int,const cpoint3&) const;

	DWORD flags;
	int sublevel;

	cpoint3 position;
	cquat	rotation;
	cpoint3 scale;

	ccolor color;
	cpoint3 min,max;//bbox

	canim anim;

	//Cast/Receive shadows
	BOOL bCastShadows;
	BOOL bReceiveShadows;

	cnode *parent;
	cnode **children;
	int nchildren;

	//User defined notes
	char *notes;//[MAX_NOTES_CHR];

	//IK
	BOOL bEnableIKDOF,bEnableIKDamp,bIKLocked;
	float ik_damp_width;
	cpoint3 ik_min_rot,ik_max_rot;//XYZ min,max rotation limits in degrees
	cquat ik_initrot;//used with DOF
	int ik_iterations,ik_recurse;

	//(ver.>=1.3.6):See through flag
	BOOL bSeeThrough;
};

//(ver.>=1.3.2)://Use this instead of the new operator and delete using cnode's Delete() func.
COREAPI cnode* CreateNewNode();

#endif //CNODE_H